/******************************************************************************
/ snooks.cpp
/
/ Copyright (c) 2017 The Human Race
/
/ Permission is hereby granted, free of charge, to any person obtaining a copy
/ of this software and associated documentation files (the "Software"), to deal
/ in the Software without restriction, including without limitation the rights to
/ use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
/ of the Software, and to permit persons to whom the Software is furnished to
/ do so, subject to the following conditions:
/
/ The above copyright notice and this permission notice shall be included in all
/ copies or substantial portions of the Software.
/
/ THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/ EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
/ OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
/ NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
/ HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
/ WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
/ FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
/ OTHER DEALINGS IN THE SOFTWARE.
/
******************************************************************************/

#include "stdafx.h"
#include "snooks.h"
#include "../Breeder/BR_Util.h"


namespace SnooksStuff {
	enum DockedState {
		kNotDocked = -1
	};
}

using namespace SnooksStuff;


void Snooks::FocusWindow(HWND win)
{
	if (win != NULL)
		SetFocus(win);
}


void Snooks::FocusMIDIEditor(COMMAND_T* ct)
{
    HWND editor = MIDIEditor_GetActive();
    if (editor != NULL) {   
	    DockWindowActivate(editor); // in case it's docked and not visible
	    FocusWindow(GetPianoView(editor));
    }
}


// Register commands - must have !WANT_LOCAL... comments!

//!WANT_LOCALIZE_1ST_STRING_BEGIN:sws_actions
static COMMAND_T g_commandTable[] =
{
	{ { DEFACCEL, "SWS/SN: Focus MIDI editor" }, "SN_FOCUS_MIDI_EDITOR", Snooks::FocusMIDIEditor, NULL },

	{ {}, LAST_COMMAND, },
};
//!WANT_LOCALIZE_1ST_STRING_END


int snooks_Init()
{
	SWSRegisterCommands(g_commandTable);
	return 1;
}
